from juno.attribute_getter import AttributeGetter
from juno.payment_object import Payment


class Charge(AttributeGetter):
    def __init__(self, attributes=None):
        if attributes is None:
            attributes = {}
        self._setattrs = []
        for key, val in attributes.items():
            if key == "payments":
                setattr(self, key, [Payment(payment_dict) for payment_dict in val])
            else:
                setattr(self, key, val)

            self._setattrs.append(key)



from juno.resources import handler_request
from juno.resources.routes import notification_routes


def find_all_event_types():
    return handler_request.get(notification_routes.get_base_url_event_types())


def create_webhook(dictionary):
    return handler_request.post(notification_routes.get_base_url_webhooks(), dictionary)


def find_all_webhooks():
    return handler_request.get(notification_routes.get_base_url_webhooks())


def find_webhook_by_id(webhook_id):
    return handler_request.get(
        notification_routes.get_specific_webhook_by_id_url(webhook_id)
    )


def update_webhook(webhook_id, dictionary):
    return handler_request.patch(
        notification_routes.get_update_webhook_url(webhook_id), dictionary
    )


def delete_webhook(webhook_id):
    return handler_request.delete(
        notification_routes.get_delete_webhook_url(webhook_id)
    )



from juno.resources import handler_request
from juno.resources.routes import payment_routes


def create(dictionary):
    return handler_request.post(payment_routes.get_base_url(), dictionary)


def refund(payment_id, dictionary={}):
    return handler_request.post(
        payment_routes.get_refund_payment_url(payment_id), dictionary
    )


def capture(payment_id, dictionary={}):
    return handler_request.post(
        payment_routes.get_capture_payment_url(payment_id), dictionary
    )




from juno.resources.handler_request import init

from juno import balance
from juno import card
from juno import charge
from juno import digital_account
from juno import notification
from juno import payment
from juno import transfer
from juno import additional_data
from juno import onboarding
from juno import plan
from juno import subscription



from datetime import datetime

from requests import Session
from requests.adapters import HTTPAdapter
from requests.auth import _basic_auth_str
from requests.packages.urllib3.util.retry import Retry

from juno import sdk


def headers():
    from .handler_request import request_authorization, KEYS

    if datetime.now() > KEYS["expires"]:
        request_authorization()

    _headers = {
        "Authorization": f"Bearer {KEYS['access_token']}",
        "X-API-Version": sdk.API_VERSION,
        "X-Resource-Token": KEYS["resource_token"],
        "Content-Type": "application/json",
    }
    return _headers


def headers_authorization():
    from .handler_request import KEYS

    _headers = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Authorization": _basic_auth_str(KEYS["client_id"], KEYS["client_secret"]),
    }
    return _headers


def requests_retry_session(
    retries=3,
    backoff_factor=0.3,
    status_forcelist=(500, 502, 504),
    session=None,
    authorization=False,
):
    session = session or Session()
    session.headers.update(headers_authorization() if authorization else headers())

    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("https://", adapter)
    return session



class ErrorResult(object):
    def __init__(self, attributes):
        self.timestamp = attributes["timestamp"]
        self.status = attributes["status"]
        self.message = attributes["error"]
        self.errors = attributes["details"]
        self.path = attributes["path"]

    def __repr__(self):
        return "<%s '%s' at %x>" % (self.__class__.__name__, self.message, id(self))

    @property
    def is_success(self):
        """ Returns whether the result from the gateway is a successful response. """

        return False



import re


def _unpack(data):
    if isinstance(data, dict):
        return data.items()
    return data


def to_snake_case(value):
    first_underscore = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", value)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", first_underscore).lower()


def keys_to_snake_case(content):
    return {to_snake_case(key): value for key, value in _unpack(content)}


def to_camel_case(value):
    content = value.split("_")
    return content[0] + "".join(
        word.capitalize() for word in content[1:] if not word.isspace()
    )


def keys_to_camel_case(content):
    return {to_camel_case(key): value for key, value in _unpack(content)}


def parse_keys(data, formatter):
    if not isinstance(data, (list, dict)):
        raise TypeError("Invalid data type, use list or dict")

    formatted = type(data)()

    is_dict = lambda x: type(x) == dict
    is_list = lambda x: type(x) == list

    for key, value in _unpack(formatter(data)):
        if is_dict(value):
            formatted[key] = parse_keys(value, formatter)
        elif is_list(value) and len(value) > 0:
            formatted[key] = []
            for val in value:
                if isinstance(val, (list, dict)):
                    val = parse_keys(val, formatter)
                formatted[key].append(val)
        else:
            formatted[key] = value
    return formatted


def camelize(data):
    return parse_keys(data, keys_to_camel_case)


def underscoreize(data):
    return parse_keys(data, keys_to_snake_case)



from juno.resources import handler_request
from juno.resources.routes import subscription_routes


def create(dictionary):
    return handler_request.post(subscription_routes.get_base_url(), dictionary)


def find_all():
    return handler_request.get(subscription_routes.get_base_url())


def find_by_id(subscription_id):
    return handler_request.get(
        subscription_routes.get_specific_subscription_by_id_url(subscription_id)
    )


def deactivation(subscription_id):
    return handler_request.post(
        subscription_routes.get_deactivation_subscription_url(subscription_id)
    )


def activation(subscription_id):
    return handler_request.post(
        subscription_routes.get_activation_subscription_url(subscription_id)
    )


def cancelation(subscription_id):
    return handler_request.post(
        subscription_routes.get_cancelation_subscription_url(subscription_id)
    )


def completion(subscription_id):
    return handler_request.post(
        subscription_routes.get_completion_subscription_url(subscription_id)
    )



from juno.resources import handler_request
from juno.resources.routes import card_routes


def tokenization(dictionary):
    return handler_request.post(card_routes.get_tokenization_card_url(), dictionary)



from juno.resources import handler_request
from juno.resources.routes import plan_routes


def create(dictionary):
    return handler_request.post(plan_routes.get_base_url(), dictionary)


def find_all():
    return handler_request.get(plan_routes.get_base_url())


def find_by_id(plan_id):
    return handler_request.get(plan_routes.get_specific_plan_by_id_url(plan_id))


def deactivation(plan_id):
    return handler_request.post(plan_routes.get_deactivation_plan_url(plan_id))


def activation(plan_id):
    return handler_request.post(plan_routes.get_activation_plan_url(plan_id))



from ..handler_request import get_resource_url


def get_base_url():
    return f"{get_resource_url()}/subscriptions"


def get_specific_subscription_by_id_url(subscription_id):
    return f"{get_base_url()}/{subscription_id}"


def get_deactivation_subscription_url(subscription_id):
    return f"{get_base_url()}/{subscription_id}/deactivation"


def get_activation_subscription_url(subscription_id):
    return f"{get_base_url()}/{subscription_id}/activation"


def get_cancelation_subscription_url(subscription_id):
    return f"{get_base_url()}/{subscription_id}/cancelation"


def get_completion_subscription_url(subscription_id):
    return f"{get_base_url()}/{subscription_id}/completion"



from juno.resources import handler_request
from juno.resources.routes import onboarding_routes


def account_new_onboarding_request(dictionary):
    return handler_request.post(onboarding_routes.get_base_url(), dictionary)


from juno.resources import handler_request
from juno.resources.routes import charge_routes


def create(dictionary):
    return handler_request.post(charge_routes.get_base_url(), dictionary)


def find_all(query_params={}):
    return handler_request.get(charge_routes.get_base_url(), query_params)


def find_by_id(charge_id):
    return handler_request.get(charge_routes.get_specific_charge_by_id_url(charge_id))


def cancelation(charge_id):
    return handler_request.put(charge_routes.get_cancelation_charge_url(charge_id))


def update_split(charge_id, dictionary):
    return handler_request.put(
        charge_routes.get_update_split_charge_url(charge_id), dictionary
    )



from juno.resources import handler_request
from juno.resources.routes import transfer_routes


def create(dictionary, resource_token=None):
    if resource_token:
        dictionary["resource_token"] = resource_token

    return handler_request.post(transfer_routes.get_base_url(), dictionary)



from ..handler_request import get_resource_url


def get_base_url():
    return f"{get_resource_url()}/notifications"


def get_base_url_event_types():
    return f"{get_base_url()}/event-types"


def get_base_url_webhooks():
    return f"{get_base_url()}/webhooks"


def get_specific_webhook_by_id_url(webhook_id):
    return f"{get_base_url_webhooks()}/{webhook_id}"


def get_update_webhook_url(webhook_id):
    return f"{get_base_url_webhooks()}/{webhook_id}"


def get_delete_webhook_url(webhook_id):
    return f"{get_base_url_webhooks()}/{webhook_id}"



from juno.resources import handler_request
from juno.resources.routes import additional_data_routes


def list_banks():
    return handler_request.get(additional_data_routes.get_banks_url())

def list_company_types():
    return handler_request.get(additional_data_routes.get_company_types_url())

def list_business_areas():
    return handler_request.get(additional_data_routes.get_business_areas_url())


from ..handler_request import get_resource_url


def get_base_url():
    return f"{get_resource_url()}/plans"


def get_specific_plan_by_id_url(plan_id):
    return f"{get_base_url()}/{plan_id}"


def get_deactivation_plan_url(plan_id):
    return f"{get_base_url()}/{plan_id}/deactivation"


def get_activation_plan_url(plan_id):
    return f"{get_base_url()}/{plan_id}/activation"



class JunoException(Exception):
    def __init__(
        self,
        message,
        timestamp=None,
        error_code=None,
        status=None,
        error=None,
        path=None,
    ):
        self.timestamp = timestamp
        self.message = message
        self.error_code = error_code
        self.status = status
        self.error = error
        self.path = path

    def __str__(self):
        if not self.error_code:
            return self.message

        return f"{self.message} (timestamp={self.timestamp}, error_code={self.error_code}, status={self.status}, error={self.error}, path={self.path})"

    def __repr__(self):
        if not self.error_code:
            return self.message

        return f"('{self.timestamp}', '{self.message}', '{self.error_code}', {self.status}, '{self.error}', '{self.path}')"


class JunoInvalidCredentials(JunoException):
    pass


class JunoNotFoundCreditCardByHash(JunoException):
    pass


class JunoCreditCardWithInsufficientBalance(JunoException):
    pass


class JunoRestrictedCreditCard(JunoException):
    pass


class JunoInvalidCreditCard(JunoException):
    pass


class JunoOperationFailed(JunoException):
    pass


class JunoCaptureValueGreaterThanAuthorizedValue(JunoException):
    pass


class JunoAlreadyRegisteredWebhookForIndicatedEvents(JunoException):
    pass



from ..handler_request import get_resource_url


def get_base_url():
    return f"{get_resource_url()}/charges"


def get_specific_charge_by_id_url(charge_id):
    return f"{get_base_url()}/{charge_id}"


def get_cancelation_charge_url(charge_id):
    return f"{get_base_url()}/{charge_id}/cancelation"


def get_update_split_charge_url(charge_id):
    return f"{get_base_url()}/{charge_id}/split"



from juno.attribute_getter import AttributeGetter


class Plan(AttributeGetter):
    def __init__(self, attributes=None):
        if attributes is None:
            attributes = {}
        self._setattrs = []
        for key, val in attributes.items():
            setattr(self, key, val)
            self._setattrs.append(key)



import re

from datetime import datetime, timedelta

from .. import exceptions
from ..charge_object import Charge
from ..error_result import ErrorResult
from ..payment_object import Payment
from ..plan_object import Plan
from ..subscription_object import Subscription
from ..successful_result import SuccessfulResult
from ..utils import camelize, underscoreize

from .requests_retry import requests_retry_session


scheme = "https"
path_authorization = "/authorization-server/oauth/token"

KEYS = {}
RESOURCE_SERVER_URL = ""
AUTHORIZATION_URL = ""

regex_charges = "^\/charges\/$"
regex_charges_detail = "^\/charges\/(.*)\/$"
regex_charges_cancelation = "^\/charges\/(.*)\/cancelation\/$"
regex_charges_split = "^\/charges\/(.*)\/split\/$"
regex_payments = "^\/payments\/$"
regex_payments_capture = "^\/payments\/(.*)\/capture\/$"
regex_payments_refunds = "^\/payments\/(.*)\/refunds\/$"
regex_plans = "^\/plans\/$"
regex_plans_detail = "^\/plans\/(.*)\/$"
regex_subscriptions = "^\/subscriptions\/$"
regex_subscriptions_detail = "^\/subscriptions\/(.*)\/$"


def data_authorization():
    _data = {"grant_type": "client_credentials"}
    return _data


def dict_to_keys(dictionary):
    global KEYS

    for key, value in dictionary.items():
        if key == "expires_in":
            KEYS["max_age"] = value
            KEYS["expires"] = datetime.now() + timedelta(seconds=value)
        else:
            KEYS[key] = value


def request_authorization():
    dict_to_keys(
        validate_authorization(
            requests_retry_session(authorization=True).post(
                AUTHORIZATION_URL, data=data_authorization()
            )
        )
    )


def validate_authorization(juno_response):
    if juno_response.status_code == 204:
        return None

    response_json = underscoreize(juno_response.json())
    if juno_response.ok:
        return response_json

    return error(response_json)


def get_data_charges(data, method):
    if "_embedded" not in data:
        return {"charges": []}

    if len(data["_embedded"]["charges"]) == 1 and method == "POST":
        return {"charge": Charge(data["_embedded"]["charges"][0])}

    return {
        "charges": [Charge(charge_dict) for charge_dict in data["_embedded"]["charges"]]
    }


def get_data_payments(data):
    if len(data["payments"]) == 1:
        return {"payment": Payment(data["payments"][0])}

    return {"payments": [Payment(payment_dict) for payment_dict in data["payments"]]}


def get_data_plans(data, method):
    if len(data["_embedded"]["plans"]) == 1 and method == "POST":
        return {"plan": Plan(data["_embedded"]["plans"][0])}

    return {"plans": [Plan(plan_dict) for plan_dict in data["_embedded"]["plans"]]}


def get_data_subscriptions(data, method):
    if len(data["_embedded"]["subscriptions"]) == 1 and method == "POST":
        return {"subscription": Subscription(data["_embedded"]["subscriptions"][0])}

    return {
        "subscriptions": [
            Subscription(subscription_dict)
            for subscription_dict in data["_embedded"]["subscriptions"]
        ]
    }


def success_result(method, url, data):
    response = data
    if (method == "GET" and re.search(regex_charges, url)) or (
        method == "POST" and re.search(regex_charges, url)
    ):
        response = get_data_charges(data, method)

    elif method == "GET" and re.search(regex_charges_detail, url):
        response = {"charge": Charge(data)}

    elif (method == "PUT" and re.search(regex_charges_cancelation, url)) or (
        method == "PUT" and re.search(regex_charges_split, url)
    ):
        return None

    elif (method == "POST" and re.search(regex_payments, url)) or (
        method == "POST" and re.search(regex_payments_capture, url)
    ):
        response = get_data_payments(data)

    elif method == "POST" and re.search(regex_payments_refunds, url):
        print("POST payments refunds")

    elif (method == "GET" and re.search(regex_plans, url)) or (
        method == "POST" and re.search(regex_plans, url)
    ):
        response = get_data_plans(data, method)

    elif method == "GET" and re.search(regex_plans_detail, url):
        response = {"plan": Plan(data)}

    elif (method == "GET" and re.search(regex_subscriptions, url)) or (
        method == "POST" and re.search(regex_subscriptions, url)
    ):
        response = get_data_subscriptions(data, method)

    elif method == "GET" and re.search(regex_subscriptions_detail, url):
        response = {"subscription": Subscription(data)}

    return SuccessfulResult(response)


def error_result(data):
    return ErrorResult(data)


def validate_response(method, end_point, juno_response):
    if juno_response.status_code == 204:
        return success_result(
            method, f'{end_point.replace(RESOURCE_SERVER_URL, "")}/', None
        )

    response_json = underscoreize(juno_response.json())

    if juno_response.ok:
        return success_result(
            method, f'{end_point.replace(RESOURCE_SERVER_URL, "")}/', response_json
        )

    return error_result(response_json)


def init(client_id=None, client_secret=None, resource_token=None, sandbox=True):
    if not client_id or not client_secret or not resource_token:
        raise exceptions.JunoInvalidCredentials("Invalid credentials")

    global KEYS, AUTHORIZATION_URL, RESOURCE_SERVER_URL
    KEYS["client_id"] = client_id
    KEYS["client_secret"] = client_secret
    KEYS["resource_token"] = resource_token

    host = "sandbox.boletobancario.com" if sandbox else "api.juno.com.br"
    path = "/api-integration" if sandbox else ""

    RESOURCE_SERVER_URL = f"{scheme}://{host}{path}"
    AUTHORIZATION_URL = f"{scheme}://{host}{path_authorization}"

    request_authorization()


def get_resource_url():
    return RESOURCE_SERVER_URL


def request_function(method):
    if method == "GET":
        return requests_retry_session().get
    elif method == "POST":
        return requests_retry_session().post
    elif method == "PUT":
        return requests_retry_session().put
    elif method == "DELETE":
        return requests_retry_session().delete
    elif method == "PATCH":
        return requests_retry_session().patch


def hook_requests(method, end_point, data):
    payload = {"json": camelize(data)}
    if method == "GET":
        payload = {"params": camelize(data)}

    juno_response = request_function(method)(end_point, **payload)

    if juno_response.status_code in [401, 403]:
        request_authorization()
        return request_function(method)(end_point, **payload)

    return juno_response


def get(end_point, data={}):
    return validate_response("GET", end_point, hook_requests("GET", end_point, data))


def post(end_point, data={}):
    return validate_response("POST", end_point, hook_requests("POST", end_point, data))


def put(end_point, data={}):
    return validate_response("PUT", end_point, hook_requests("PUT", end_point, data))


def delete(end_point, data={}):
    return validate_response(
        "DELETE", end_point, hook_requests("DELETE", end_point, data)
    )


def patch(end_point, data={}):
    return validate_response("PATCH", end_point, hook_requests("PATCH", end_point, data))


def error(data):
    if "details" in data:
        raise_exception_from_error_code(
            data["details"][0]["error_code"], data["details"][0]["message"]
        )

    raise Exception(data)


capture_value_greater_than_authorized_value_error_code = "289999"
capture_value_greater_than_authorized_value_message = (
    "O valor informado na captura é maior que o valor autorizado"
)

not_found_credit_card_by_hash_error_code = "501008"

operation_failed_error_code = "373014"

invalid_credit_card_error_code = "289999"
invalid_credit_card_message = "Não autorizado. Cartão inválido."

restricted_credit_card_error_code = "289999"
restricted_credit_card_message = None

credit_card_with_insufficient_balance_error_code = "289999"
credit_card_with_insufficient_balance_message = "Não autorizado. Saldo insuficiente."

already_registered_webhook_for_indicated_events_error_code = "441009"


def raise_exception_from_error_code(error_code, message=None):
    if error_code == not_found_credit_card_by_hash_error_code:
        raise exceptions.JunoNotFoundCreditCardByHash(message)
    elif (
        error_code == credit_card_with_insufficient_balance_error_code
        and message == credit_card_with_insufficient_balance_message
    ):
        raise exceptions.JunoCreditCardWithInsufficientBalance(message)
    elif (
        error_code == restricted_credit_card_error_code
        and message == restricted_credit_card_message
    ):
        raise exceptions.JunoRestrictedCreditCard(message)
    elif (
        error_code == invalid_credit_card_error_code
        and message == invalid_credit_card_message
    ):
        raise exceptions.JunoInvalidCreditCard(message)
    elif error_code == operation_failed_error_code:
        raise exceptions.JunoOperationFailed(message)
    elif (
        error_code == capture_value_greater_than_authorized_value_error_code
        and message == capture_value_greater_than_authorized_value_message
    ):
        raise exceptions.JunoCaptureValueGreaterThanAuthorizedValue(message)
    elif error_code == already_registered_webhook_for_indicated_events_error_code:
        raise exceptions.JunoAlreadyRegisteredWebhookForIndicatedEvents(message)
    else:
        raise exceptions.JunoException(message)



from ..handler_request import get_resource_url


def get_base_url():
    return f"{get_resource_url()}/payments"


def get_refund_payment_url(payment_id):
    return f"{get_base_url()}/{payment_id}/refunds"


def get_capture_payment_url(payment_id):
    return f"{get_base_url()}/{payment_id}/capture"



from ..handler_request import get_resource_url


def get_base_url():
    return f"{get_resource_url()}/data"

def get_banks_url():
    return f"{get_base_url()}/banks"

def get_company_types_url():
    return f"{get_base_url()}/company-types"

def get_business_areas_url():
    return f"{get_base_url()}/business-areas"


from setuptools import setup, find_packages


__description__ = "Juno Python"
with open("README.md", "r") as fh:
    __long_description__ = fh.read()

__author__ = "Manaia Junior"
__author_email__ = "manaiajr.23@gmail.com"

setup(
    name="juno-python",
    version="0.4.0",
    url="https://github.com/mjr/juno-python",
    author=__author__,
    author_email=__author_email__,
    license="MIT",
    description=__description__,
    long_description=__long_description__,
    long_description_content_type="text/markdown",
    keywords="Payment, Juno",
    packages=find_packages(),
    install_requires=["requests>=2.24.0"],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
)



from juno.attribute_getter import AttributeGetter


class SuccessfulResult(AttributeGetter):
    @property
    def is_success(self):
        """ Returns whether the result from the gateway is a successful response. """
        return True



from juno.resources import handler_request
from juno.resources.routes import balance_routes


def find_by_resource_token(resource_token=None):
    return handler_request.get(
        balance_routes.get_base_url(),
        {"resource_token": resource_token} if resource_token else {},
    )
